"""
Simple network utilities


## TODO
create a 'get_headers' method to use a headless-browser to make a call,
and return all the request headers.

"""
from requests import Response, RequestException

from .simple_utils import *
from sgrequests import SgRequests
import bs4
import urllib.parse
from bs4 import BeautifulSoup
from requests_toolbelt.utils import dump
import sglogging

log = sglogging.SgLogSetup().get_logger('simple_network_utils')

def fetch_json(request_url: str,
               path_to_locations: List[str] = [],
               query_params: dict = {},
               data_params: dict = {},
               headers: dict = {},
               retries: int = 10,
               retry_backoff_factor=0.3) -> List[dict]:
    """
    Fetches locations json, and returns as a list of parsed dictionaries.
    Assumes a regular structure with root element, and sibling-elements.

    :param path_to_locations: the path, in the json, to the locations array.

    :param request_url: See `fetch_data`
    :param query_params: See `fetch_data`
    :param data_params: See `fetch_data`
    :param headers: See `fetch_data`
    :param retries: See `fetch_data`
    :param retry_backoff_factor: See `fetch_data`

    :return: a list of parsed dictionaries
    """

    response = fetch_data(request_url=request_url,
                          data_params=data_params,
                          query_params=query_params,
                          headers=headers,
                          retries=retries,
                          retry_backoff_factor=retry_backoff_factor)

    json_result = response.json()

    return [drill_down_into(json_result, path_to_locations)]

def fetch_xml(request_url: str,
              root_node_name: str,
              location_node_name: str,
              location_parser: Callable[[bs4.Tag], dict] = xml_to_dict,
              location_node_properties: Dict[str, str] = {},
              query_params: dict = {},
              data_params: dict = {},
              headers: dict = {},
              xml_parser: str = 'lxml',
              retries: int = 10,
              retry_backoff_factor = 0.3) -> List[dict]:
    """
    Fetches locations xml, and returns as a list of parsed dictionaries.
    Assumes a regular structure with root element, and sibling-elements.

    :param root_node_name: root element under which locations are stored.
    :param location_node_name: element name of each location.
    :param location_node_properties: the properties of the node, e.g. {'class':'abc def'}
    :param location_parser: function that converts a location Tag into a dict.
    :param request_url: See `fetch_data`
    :param query_params: See `fetch_data`
    :param data_params: See `fetch_data`
    :param headers: See `fetch_data`
    :param xml_parser: The xml parser used by BeautifulSoup
    :param retries: See `fetch_data`
    :param retry_backoff_factor: See `fetch_data`

    :return: a generator of parsed dictionaries
    """

    response = fetch_data(request_url=request_url,
                          data_params=data_params,
                          query_params=query_params,
                          headers=headers,
                          retries=retries,
                          retry_backoff_factor=retry_backoff_factor)

    xml_result = BeautifulSoup(response.text, xml_parser)

    root_node = xml_result.find(root_node_name)

    location_nodes = root_node.find_all(location_node_name, location_node_properties)

    for location in location_nodes:
        yield location_parser(location)

def fetch_data(request_url: str,
               query_params: dict = {},
               data_params: dict = {},
               headers: dict = {},
               retries: int = 10,
               retry_backoff_factor=0.3) -> Response:
    """
    Returns the `SgRequests` response if the response status code is in (200, 299)

    :param request_url: The URL to fetch.
    :param query_params: The raw query params, unquoted and unescaped.
    :param data_params: The raw form data.
    :param headers: The headers.
    :param retries: How many retries to attempt.
    :param retry_backoff_factor: The backoff factor for the retries; same as in SgRequests.
    :return:
    """

    session = SgRequests().requests_retry_session(retries=retries, backoff_factor=retry_backoff_factor)

    response = session.get(request_url, data=urllib.parse.urlencode(data_params), params=urllib.parse.urlencode(query_params), headers=headers)

    if response.status_code < 200 or response.status_code > 299:
        log.error("API call is not successful; result status: " + str(response.status_code))
        log.error(dump.dump_all(response).decode("utf-8"))
        raise RequestException(f'Bad status code: {response.status_code}', response=response)

    return response

def paginated(fetch_results: Callable[[int], list], max_per_page: int, first_page: int = 1) -> list:
    """
    Simple utility to fetch paginated results.

    :param fetch_results: Given the page number, fetch the list (or Generator) of results from that page.
    :param max_per_page: The expected cap of results per page; if the count falls below it, the function will terminate.
    :param first_page: What's the first page? Defaults to 1. Page numbers will be incremented from this one.
    :return: The generator of results from all pages.
    """
    keep_going = True
    page_num = first_page
    while keep_going:
        result_count = 0
        for result in fetch_results(page_num):
            result_count += 1
            yield result

        if result_count < max_per_page:
            keep_going = False
        else:
            page_num += 1
